--  Distributed Haskell: ports library
--
--  Author : Manuel M. T. Chakravarty
--  Created: 20 March 2000
--
--  Version $Revision: 1.12 $ from $Date: 2001/07/02 07:39:46 $
--
--  Copyright (c) [2000..2001] Manuel M. T. Chakravarty
--
--  This file is free software; you can redistribute it and/or modify
--  it under the terms of the GNU General Public License as published by
--  the Free Software Foundation; either version 2 of the License, or
--  (at your option) any later version.
--
--  This file is distributed in the hope that it will be useful,
--  but WITHOUT ANY WARRANTY; without even the implied warranty of
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--  GNU General Public License for more details.
--
--- DESCRIPTION ---------------------------------------------------------------
--
--  Provides ports in the spirit of 
--
--    Distributed Haskell: Goffin on the Internet. Manuel M. T. Chakravarty,
--    Yike Guo, and Martin Khler. In M. Sato and Y. Toyama, editors,
--    Proceedings of the Third Fuji International Symposium on Functional and
--    Logic Programming, World Scientific Publishers, pp 80-97, 1998.
--
--  plus, by now, a considerable amount of extras.
--
--- DOCU ----------------------------------------------------------------------
--
--  language: Haskell 98 + Concurrency + Weak pointers 
--
--  Structure of Ports
--  ~~~~~~~~~~~~~~~~~~
--  A port consists of two components:
--  (1) the port data and
--  (2) the port filter.
--
--  The port filter is a function that censors any new port value (and may
--  alter the value or discard it altogether).  The port data consists of the
--  current port value plus the hole in the port stream that has to be
--  instantiated next; furthermore, contains a port notifier, which is invoked
--  on every new value put into the out stream of the port.  The port data is
--  stored in an `MVar' to allow port updates and perform these updates
--  atomically.  The later is achieved by taking the port data out of the
--  `MVar' when a critical section in the update code is reached and putting
--  it back at the end of the critical section.
--
--  As soon as a port is no dead (and this is spotted by the storage manager),
--  it is automagically closed.  This is sometimes essential and sometimes
--  just convenient, but care has to be taken, as space leaks can lead to
--  deadlock. 
--
--  Port Streams
--  ~~~~~~~~~~~~
--  Port streams are represented as lists with a hole, ie, lists terminated by
--  a single assignment variable, which is updated to extended the stream.
--  This is a standard technique from logic programming.
--
--  Proxy Ports
--  ~~~~~~~~~~~
--  A port may serve as a proxy for a stateful object, such as a GUI widget.
--  In this case, the proxy mirrors the (externally visible) state of the
--  object and takes care to avoid feedback loops between the port and the
--  object.  Moreover, port notifiers can be useful in other cases where a
--  synchronous notification of changes of a port value are required (eg, when
--  port values are dumped to stdout and interleaving with other output should
--  be minimised).
--
--  Internal port notifier
--  ~~~~~~~~~~~~~~~~~~~~~~
--  Each port has an internal port notifier that is run on the filtered input
--  messages.  The notifier can be modified (extended) over the lifetime of a
--  port.  As notifications are only send out on filtered messages, the
--  message will effectively be filtered twice for each linked  port (and it
--  won't arrive at any linked port at all if it is already discarded by the
--  filter of the port receiving the original message).  Internally, messages
--  are tagged as either coming from the outside, from a linked port, or from
--  an object for which the current port is a proxy.  These tags are used to
--  prevent feedback loops.
--
--  Note that a port is proxy to at most one object; thus, there is also at
--  most one notifier for such an object.
--
--  Concurrency Control
--  ~~~~~~~~~~~~~~~~~~~
--  `Concurrent.yield' is important to ensure that listeners get a chance soon
--  after a port is updated (even if the routine that did the update goes into
--  non-preemptive non-Haskell land).
--
--  All port primitives avoid to block indefinitely.  They use blocking
--  operations only to realise mutual exclusion to critical sections in which
--  no operation remains for longer periods of time.  However, they may block
--  of course when any of the supplied arguments that needs to be touched is
--  not yet available.
--
--  Port primitives do not spawn asynchronous operations to complete their
--  task - ie, when a port primtive returns it's operation has been fully
--  executed.  In particular, the primitive will not access the port anymore
--  after it returned, and so, will not be affected if the port is closed.
--
--- TODO ----------------------------------------------------------------------
--
--  * BUG: Port linking doesn't work properly for more than two ports.  When a
--	   port is linked to a port that is already linked to a third port,
--	   the newly linked port has to be registered in the notifiers of
--	   *both* ports.
--
--  * Problem with finalisers: There can be a deadlock if all threads wait for
--      finalisers to run (so that unused ports get closed); forcing a GC in
--      this situation helps, but IMO this should be done automagically by the
--      RTS. [Should be ok from GHC 5.02 onwards.]
--
--  * All critical sections should probably be protected against exceptions.
--    The port internal MVar has to be put back even if an exception occurs
--    during the critical section (eg, an async exc); otherwise, the system
--    may deadlock.
--

module Ports (
  PortFilter, PortNotifier, Port, 
  withPorts, newPort, listenToNewPort, newFilteringPort, newProxyPort,
  closePort, isClosedPort, waitUntilClosedPort, keepAlivePort, listenToPort,
  (<--), (<==), (<-$), chainPorts, linkPorts, (<->), peekIntoPort,

  -- thread routines
  --
  (&), keepAliveForkIO,

  -- re-exported
  --
  forkIO
) where

import Prelude     hiding (catch)
import Maybe
import Monad	   (unless, when)
import System	   (getArgs)

import Concurrent  (ThreadId, myThreadId, 
		    MVar, newMVar, newEmptyMVar, takeMVar, putMVar,
		    readMVar, swapMVar, isEmptyMVar, forkIO, yield,
		    threadDelay) 
import Exception   (throw, throwTo, catch, finally)
import IOExts      (unsafePerformIO, performGC)  -- only temporary
import StablePtr   (newStablePtr, freeStablePtr) -- only temporary
import Weak        (addFinalizer)

import PortsConfig (debug)
import SVars	   (V, newV, valV, (<<))


infixl 2 &
infix  3 <--, <==, <-$


-- ports data structures
-- ---------------------

-- port filter are used to screen messages before they are allowed into the
-- port (EXPORTED)
--
-- * messages may be rejected (by returning `Nothing')
--
-- * and may be altered (by returning the altered value)
--
-- * the first argument is the current value of the port and the second is the 
--   value of the current message
--
type PortFilter a = a -> a -> Maybe a

-- port notifier (EXPORTED)
--
-- * an action either notifying a port of a new message or notifying some
--   other stateful object (eg, a widget) of a new port value
--
type PortNotifier a = a -> IO ()

-- we have three different kinds of messages
--
data MsgKind = ExternalMsg	-- from an external source (normal send)
	     | BuddyMsg		-- from a buddy, ie, linked port
	     | ProxyMsg		-- from an object for which we are a proxy
	     deriving (Eq)

-- data specifying the current state of a port
--
type PortData a = (a,				-- current port value
		   V [a],			-- current hole in the stream
		   MsgKind -> PortNotifier a)	-- internal update notification

-- ports structure (EXPORTED ABSTRACTLY)
--
-- * the `MVar' is empty while a port operation is in a critical section
--
-- * the `MVar' will contain `Nothing' after the port was closed
--
data Port a = Port (MVar (Maybe (PortData a)))	-- updatable port data
		   (PortFilter a)		-- screens incoming messages

-- list of synchronisation points that have to be honoured before terminating
--
terminationSyncPoints :: MVar [MVar ()]
{-# NOINLINE terminationSyncPoints #-}
terminationSyncPoints = unsafePerformIO $ newMVar []


-- ports initilisation
-- -------------------

-- initialise ports environment
--
-- * lists remotely callable entry points
--
-- * this function will not terminate unless all `keep alive' ports are closed
--
withPorts          :: [(String, IO ())] -> ([String] -> IO ()) -> IO ()
withPorts rep main  =
  do
    unless (null rep) $
      putStrLn "Warning: Distribution not supported yet!"
    args <- getArgs
    main args 
    takeMVar terminationSyncPoints >>= mapM_ takeMVar  -- wait for sync points
    -- FIXME: This is subtly wrong!  
    --	      If after the `terminationSyncPoints' have been taken, any other
    --	      thread calls `keepAlivePort', that thread will block on the
    --	      empty `terminationSyncPoints'.


-- ports core functionality
-- ------------------------

-- create a new port (EXPORTED)
--
-- * the argument is the intial values used in case an application of a
--   state transformer is performed before the first message is sent; it,
--   however, never appears in any output stream
--
-- * the port is automagically closed when it is garbage collected (and not
--   claosed yet)
--
newPort   :: a -> IO (Port a)
newPort x  = newFilteringPort x (const Just)    -- trivial filter

-- create a new port and obtain its message stream (EXPORTED)
--
listenToNewPort   :: a -> IO (Port a, [a])
listenToNewPort x  =
  do
    p  <- newPort x
    ms <- listenToPort p
    return (p, ms)

-- create a new port that has a port filter (EXPORTED)
--
-- * the argument is the intial value used in case an application of a
--   state transformer is performed before the first message is sent; it,
--   however, never appears in any output stream
--
-- * the port is automagically closed when it is garbage collected (and not
--   closed yet)
--
-- * CONSTRAINT: The port filter must not include blocking operations.  It is
--		 evaluated in a critical section that manipulates the internal
--		 representation of ports.  If it blocks, deadlocks may occur.
--
newFilteringPort     :: a -> PortFilter a -> IO (Port a)
newFilteringPort x f  = setupPort x f (const nop)

-- create a new proxy port (EXPORTED)
--
-- * the port notifier passed as an argument is used to update the `real'
--   object about a change of the port value
--
-- * in contrast, the port notifier returned as a result must be used to
--   notify the proxy of a change of the state of the `real' object
--
-- * it is ensured that no feedback loop is generated
--
-- * the port is automagically closed when it is garbage collected (and not
--   closed yet)
--
newProxyPort      :: a -> PortNotifier a -> IO (Port a, PortNotifier a)
newProxyPort x pn  = 
  do
    let pn' msgType = if (msgType /= ProxyMsg)  -- don't feedback proxy msgs
		      then pn 
		      else nop
    p <- setupPort x (const Just) pn'
    return $ (p, insertMessage p ProxyMsg)

-- close the given port (EXPORTED)
--
-- * this closes the stream and empties the `MVar' to signal other attempts on 
--   accessing this port that it is closed
--
closePort             :: Port a -> IO ()
closePort (Port mv _)  = 
  do
    portData <- swapMVar mv Nothing -- the port will definitely be closed now
    case portData of
      Just (_, hole, _) -> do
			     hole << []		-- close the stream
			     yield
      Nothing           -> putStr "Warning: Ports.closePort: Attempted to \ 
			          \close already closed port!\n"

-- check whether a given port is closed (EXPORTED)
--
isClosedPort		 :: Port a -> IO Bool
isClosedPort (Port mv _)  =
  do
    portData <- readMVar mv
    return (isNothing portData)

-- suspend until the given port is closed (EXPORTED)
--
waitUntilClosedPort   :: Port a -> IO ()
--
-- FIXME: Is it worthwhile to have a more sophisticated implementation that
--	  avoids walking the stream.
--
waitUntilClosedPort p  = 
  do
    stream <- listenToPort p
    walk stream
  where
    walk []     = return ()
    walk (_:xs) = walk xs

-- make sure the program does not terminate before the given port is closed
-- (EXPORTED)
--
keepAlivePort   :: Port a -> IO ()
keepAlivePort p  = 
  do
    sp  <- newEmptyMVar
    sps <- takeMVar terminationSyncPoints
    putMVar terminationSyncPoints (sp:sps)
    addFinalizer p $ putMVar sp ()

-- attaches a stream to a port that collects the messages it (EXPORTED)
--
-- * the stream starts to collect the messages from the moment it is attached
--   - ie, it does not see old messages
--
listenToPort             :: Port a -> IO [a]
listenToPort (Port mv _)  = 
  do
    portData <- readMVar mv
    case portData of
      Just (_, hole, _) -> return (valV hole)
      Nothing           -> do
			     putStr "Warning: Ports.listenToPort: Attempted \ 
				    \to listen to a closed port!\n"
			     return []

-- send a message to a port (EXPORTED)
--
(<--)               :: Port a -> a -> IO ()
p@(Port mv f) <-- m  = insertMessage p ExternalMsg m

-- send a list of messages to a port (EXPORTED)
--
-- * the current thread *will* block if the evaluation of the second
--   argument blocks (eg, if it is a stream)
--
-- * there is no guarantee that the messages are not interleaved with
--   concurrently sent messages
--
(<==)    :: Port a -> [a] -> IO ()
p <== ms  = do
  mapM_ (p <--) ms
  return ()

-- send a state transformer to a port (EXPORTED)
--
-- * it will be applied to the message most recently seen on the port; the
--   result is send as a message to the port
--
-- * CONSTRAINT: The state transformer must not include blocking operations.
--		 It is evaluated in a critical section that manipulates the
--		 internal representation of ports.  If it blocks, deadlocks
--		 may occur.
--
(<-$)             :: Port a -> (a -> a) -> IO ()
--
-- Note that the seemingly equivalent peeking into the port, applying the
-- state transformer, and then, sending the result to the port is a much
-- weaker operation.  The following code ensures that if the transformed state
-- makes it into the port, it will be the value right after the argument
-- passed to the state transformer - the variant using `peekIntoPort' cannot
-- give this guarantee.
--
(Port mv f) <-$ g  = 
  do
    onDebugLogEmpty mv "(<-$) blocks"
    portData <- takeMVar mv
    case portData of
      Nothing                  -> do
				    putMVar mv Nothing
				    putStr "Warning: (Ports.<-$): Attempted \ 
					   \to transform a closed port!\n"
      Just state@(x, hole, pn) ->		   -- current end of stream
        case f x (g x) of
          Nothing -> putMVar mv $ Just state	   -- no change
	  Just m' -> do
		       newHole <- newV	           -- new end of stream
		       hole << m' : (valV newHole) -- extend stream
		       putMVar mv $ Just (m', newHole, pn)
					           -- remember new end
		       pn ExternalMsg m'	   -- invoke notifier
		       yield			   -- any runnable threads?

-- chain two ports (EXPORTED)
--
-- * the output of one port is sent to the second after being filtered through
--   a translation function
--
chainPorts            :: (a -> b) -> Port a -> Port b -> IO ()
chainPorts aToB pa pb  = do
  stream <- listenToPort pa
  pb <== map aToB stream

-- link two ports with translation functions (EXPORTED)
--
-- * any message to one port will - after translation - also be send to the
--   other one
--
linkPorts                 :: (a -> b) -> (b -> a) -> Port a -> Port b -> IO ()
linkPorts aToB bToA pa pb  =
  do
    addNotifier pa (insertMessage pb BuddyMsg . aToB)
    addNotifier pb (insertMessage pa BuddyMsg . bToA)

-- link two ports (EXPORTED)
--
-- * any message to one port will also be send to the other one
--
(<->) :: Port a -> Port a -> IO ()
(<->)  = linkPorts id id

-- observe the momentary current value of the port (EXPORTED)
--
-- * returns `Nothing' if the port is closed
--
peekIntoPort             :: Port a -> IO (Maybe a)
peekIntoPort (Port mv f)  = 
  do
    portData <- readMVar mv
    case portData of
      Nothing        -> return Nothing
      Just (x, _, _) -> return $ Just x


-- thread routines
-- ---------------
-- FIXME: They should go into an extra module

-- fork-join concurrency (EXPORTED)
--
-- * exception safe: when any of the two threads terminates on an exception,
--   it is guaranteed that the same exception is raised in the main thread
--
(&)   :: IO a -> IO b -> IO (a, b)
m & n  = do
           mainId  <- myThreadId
	   mResVar <- newEmptyMVar
	   forkIO $ (m >>= putMVar mResVar)
		    `catch` throwTo mainId	-- propagate exceptions
	   nRes <- n
	   mRes <- takeMVar mResVar
	   return (mRes, nRes)

-- make sure the program does not terminate before the current thread
-- terminates (EXPORTED)
--
keepAliveForkIO   :: IO () -> IO ThreadId
keepAliveForkIO m  = 
  do
    sp  <- newEmptyMVar
    sps <- takeMVar terminationSyncPoints
    putMVar terminationSyncPoints (sp:sps)
    forkIO $ m `finally` putMVar sp ()


-- auxiliary routines
-- ------------------

-- no notifier
--
nop   :: PortNotifier a
nop _  = return ()

-- creates a port structure
--
-- * the argument is the intial value used in case an application of a
--   state transformer is performed before the first message is sent; it,
--   however, never appears in any output stream
--
-- * the port is automagically closed when it is garbage collected (and not
--   closed yet)
--
-- * CONSTRAINT: The port filter must not include blocking operations.  It is
--		 evaluated in a critical section that manipulates the internal
--		 representation of ports.  If it blocks, deadlocks may occur.
--
setupPort :: a -> PortFilter a -> (MsgKind -> PortNotifier a) -> IO (Port a)
setupPort x f pn  =
  do
    hole <- newV
    mv <- newMVar $ Just (x, hole, pn)
    let p = Port mv f
    -- FIXME: Kludge until RTS is fixed! 
    --        (makes it explicit to GC that mv can still be written)
    stableMV <- newStablePtr mv
    -- End of Kludge
    addFinalizer p (ensurePortIsClosed p stableMV)
    -- FIXME: Evil Kludge until RTS is fixed!
    ensureRegularGC
    -- End of Evil Kludge
    return p
  where
    ensurePortIsClosed p stableMV = do
      closed <- isClosedPort p
      logDebug $ "Ports: Finaliser: port already closed: " 
		 ++ show closed
      unless closed $ 
        closePort p
      -- FIXME: Kludge until RTS is fixed! 
      freeStablePtr stableMV
      -- End of Kludge

-- FIXME: Evil Kludge until RTS is fixed!
sentinel :: MVar Bool
{-# NOINLINE sentinel #-}
sentinel = unsafePerformIO $ newMVar False  -- GC thread started?
ensureRegularGC = do
    started <- swapMVar sentinel True
    unless started $ do
      forkIODebug "GC trigger thread" $ 
        let gc = threadDelay 100000 >> performGC >> gc in gc
      return ()
-- End of Evil Kludge

-- pass a message to a port ignoring the buddy list when the first argument is
-- `false'
--
insertMessage                       :: Port a -> MsgKind -> a -> IO ()
insertMessage (Port mv f) msgKind m  =
  do
    onDebugLogEmpty mv "insertMessage blocks"
    portData <- takeMVar mv
    case portData of
      Nothing                  -> do
				    putMVar mv Nothing
				    putStr warnMsg
      Just state@(x, hole, pn) ->		   -- current end of stream
       do
	case f x m of
	  Nothing -> putMVar mv $ Just state	   -- no change
	  Just m' -> do
		       newHole <- newV		   -- create new end of stream
		       hole << m' : (valV newHole) -- extend stream
		       putMVar mv $ Just (m', newHole, pn)
						   -- store new end of stream
		       pn msgKind m'		   -- invoke notifier
		       yield			   -- any runnable threads?
  where
    warnMsg = "Warning: (Ports.insertMessage): Attempted \ 
	      \to write to a closed port!\n"

-- add a notifier to the buddy list
--
addNotifier                 :: Port a -> PortNotifier a -> IO ()
addNotifier (Port mv f) pn2  =
  do
    onDebugLogEmpty mv "addNotifier blocks"
    portData <- takeMVar mv
    case portData of
      Nothing             -> do
			       putMVar mv Nothing
			       putStr warnMsg
      Just (x, hole, pn1) -> do
			       let pn' msgType x = 
				     do
				       pn1 msgType x
				       when (msgType /= BuddyMsg) $ 
					 pn2 x
			       putMVar mv $ Just (x, hole, pn')
  where
    warnMsg = "Warning: (Ports.addNotifier): Attempted \ 
	      \to modify a closed port!\n"


-- auxilliary routines for debugging
-- ---------------------------------
-- FIXME: these routines are duplicated in `Processes'; maybe introduce a
--	  `PortsDebug' module

-- a variant of `forkIO' that installs an exception sentry if compiled with
-- debugging enabled
--
forkIODebug          :: String -> IO () -> IO ThreadId
forkIODebug desc com  = forkIO $ excSentry desc com

-- report exceptions caught by the given computation
--
-- * only in effect when compiled with debugging enabled
--
excSentry :: String	-- thread descriptor
	  -> IO ()	-- computation to watch
	  -> IO ()
{-# INLINE excSentry #-}
excSentry desc com | not debug = com
		   | otherwise = com `catch` logExc
  where
    logExc exc = do
		   putStrLn $ "Processes: Thread \"" ++ desc ++ "\": "
			      ++ show exc
		   throw exc

-- execute only if debugging on
--
onDebug                  :: IO () -> IO ()
{-# INLINE onDebug #-}
onDebug com  | debug      = com
	     | otherwise  = return ()

-- emit debugging message
--
logDebug                 :: String -> IO ()
{-# INLINE logDebug #-}
logDebug msg = onDebug $ putStrLn msg

-- if debugging switched on, write a log message if the `MVar' is empty
--
onDebugLogEmpty        :: MVar a -> String -> IO ()
onDebugLogEmpty mv msg  = 
  onDebug $ do
    isEmpty <- isEmptyMVar mv
    when isEmpty $
      putStrLn $ msg ++ ": empty `MVar'"
