--  Haskell Ports Library: Process management
--
--  Author : Manuel M. T. Chakravarty
--  Created: 14 May 2001
--
--  Version $Revision: 1.7 $ from $Date: 2001/07/04 16:15:52 $
--
--  Copyright (c) 2001 Manuel M. T. Chakravarty
--
--  This file is free software; you can redistribute it and/or modify
--  it under the terms of the GNU General Public License as published by
--  the Free Software Foundation; either version 2 of the License, or
--  (at your option) any later version.
--
--  This file is distributed in the hope that it will be useful,
--  but WITHOUT ANY WARRANTY; without even the implied warranty of
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--  GNU General Public License for more details.
--
--- DESCRIPTION ---------------------------------------------------------------
--
--  This modules provides support for handling child processes and their input
--  and output streams.
--
--- DOCU ----------------------------------------------------------------------
--
--  language: Haskell 98 + HsLibs POSIX library + Concurrency + Exceptions
--
--  Process Proxies
--  ~~~~~~~~~~~~~~~
--
--  Any child process is associated with a Haskell thread called its *process
--  proxy*.  The process proxy implements the conversion between streams (ie,
--  lazy Haskell character lists) and the OS streams represented by the file
--  descriptors for stdin, stdout & stderr.  Moreover, the proxy acts as a
--  supervisor to the process and as such handles termination of the process
--  etc.  Internally, the proxy may use a number of threads to implemented the
--  overall proxy functionality.
--
--  I/O & Threads
--  ~~~~~~~~~~~~~
--
--  Currently, we rely on standard handle-based I/O to be thread friendly, ie,
--  only to block individual threads and not the whole Haskell runtime.  If
--  this turns out to impose some overhead, we can always switch to directly
--  operate on Posix file descriptors.
--
--- TODO ----------------------------------------------------------------------
--
--  * `String' for representing input and output streams may not be a good
--    choice in the presence of Unicode support; but Word8 seems to be a bit
--    unwieldy...
--
--  * possible error conditions have to be handled better
--
--  * Benchmarking: how much throughput can we get with a char-by-char copy;
--    how much better is reading in blocks and using `<==' to send; how much
--    better would a `Port String' be?
--
--  * An alternative design: instead of using streams for stdin, stdout, and
--    stderr, a process provides a port for each.  Piping one into another
--    would, then, be simple port linking.  Same for merging stderr streams.
--

module Processes (
  Proc, ProcId, toProc, proc, (>|), procStatus, procKill
) where

import Prelude     hiding (catch)
import Directory   (setCurrentDirectory)
import IO	   (Handle, hClose, hPutChar, hGetChar, isEOFError)
import Maybe       (isJust)
import Monad       (unless, when)
import System	   (ExitCode(..))

import Concurrent  (ThreadId, forkIO, mergeIO, threadDelay)
import Exception   (throw, catch, finally)
import Posix       (ProcessID, FdOption(..), ProcessStatus(..), forkProcess,
		    executeFile, createPipe, handleToFd, fdToHandle, dupTo,
		    intToFd, setFdOption, getProcessStatus, getErrorCode,
		    noChildProcess, exitImmediately, signalProcess,
		    killProcess)

import PortsConfig
import SVars       (V, newV, valV, (<<))
import Ports       (Port, newPort, closePort, listenToPort, (<--), (<==))


infixl 1 >|


-- processes & process creation
-- ----------------------------

-- a process converts a stdin stream into a process id, a stdout stream, and a
-- stderr stream (EXPORTED)
--
type Proc =  String        -- stdin
	  -> Port Char     -- stdout
	  -> Port Char     -- stderr
	  -> IO ProcId

-- abstract process handle (EXPORTED ABSTRACTLY)
--
-- * the unit in `NoProcId' and the exit code in `ProcId' are implicit
--   synchronisation points; when they are available the process has
--   completed (and won't touch any ports anymore)
--
data ProcId = NoProcId ()
	    | ProcId   ProcessID [ThreadId] ExitCode
	    | PairId   ProcId ProcId

-- turn a stream-processor function into a process (EXPORTED)
--
-- * the stream processor is executed asynchronously
--
toProc   :: (String -> String) -> Proc
toProc f  = \stdin stdout _ -> do
                                 sync <- newV
			         forkIODebug "toProc" $ (stdout <== f stdin)
						        `finally` (sync << ())
				 return $ NoProcId (valV sync)

-- execute a sub-process asynchronously (EXPORTED)
--
-- * the given executable is fork()ed and its standard input and output
--   streams redirected to lazy character lists
--
-- * the executable is searched for in the current PATH
--
proc :: FilePath			-- name of executable
     -> [String]			-- arguments
     -> Proc
proc fname args = \stdin stdout stderr -> do
  logDebug $ "Processes.proc: Setting up pipes for `" ++ fname ++ "'"
  --
  -- we need three pipes to re-direct the standard fds of the child
  --
  (stdinReadFD , stdinWriteFD ) <- createPipe
  (stdoutReadFD, stdoutWriteFD) <- createPipe
  (stderrReadFD, stderrWriteFD) <- createPipe
  stdinReadHandle   <- fdToHandle stdinReadFD
  stdinWriteHandle  <- fdToHandle stdinWriteFD
  stdoutReadHandle  <- fdToHandle stdoutReadFD
  stdoutWriteHandle <- fdToHandle stdoutWriteFD
  stderrReadHandle  <- fdToHandle stderrReadFD
  stderrWriteHandle <- fdToHandle stderrWriteFD
  --
  -- the child must close the pipe ends that it doesn't use (especially, the
  -- write end)
  --
  setFdOption stdinWriteFD CloseOnExec True
  setFdOption stdoutReadFD CloseOnExec True
  setFdOption stderrReadFD CloseOnExec True
  --
  -- start child process
  --
  logDebug $ "Processes.proc: Fork()ing `" ++ fname ++ "'"
  pid <- runProcess fname args Nothing Nothing 
		    (Just stdinReadHandle) 
		    (Just stdoutWriteHandle) (Just stderrWriteHandle)
  logDebug $ "Processes.proc: Fork()ed `" ++ fname ++ "'"
  --
  -- make sure that we don't block on the pipes and close the unused ends
  --
  setFdOption stdinWriteFD NonBlockingRead True
  setFdOption stdoutReadFD NonBlockingRead True
  setFdOption stderrReadFD NonBlockingRead True
  hClose stdinReadHandle
  hClose stdoutWriteHandle
  hClose stderrWriteHandle
  --
  -- start the process proxies
  --
  logDebug $ "Processes.proc: Starting proxies for `" ++ fname ++ "'"
  stdinProxy  <- forkIODebug stdinDesc  $ writer stdinWriteHandle stdin
--  stdoutProxy <- forkIODebug stdoutDesc $ reader stdoutReadHandle stdout
--  stderrProxy <- forkIODebug stderrDesc $ reader stderrReadHandle stderr
-- FIXME: for some strange reason, the stdout port is sometimes finalised to
--   late, something must be hanging on to it...
  stdoutProxy <- forkIODebug stdoutDesc $ reader stdoutReadHandle stdout True
  stderrProxy <- forkIODebug stderrDesc $ reader stderrReadHandle stderr True
  --
  -- obtain the process status
  --
  vec <- newV
  forkIODebug ecDesc $ waitForExitCode pid vec
--  return $ ProcId pid [stdinProxy, stdoutProxy, stderrProxy] (valV vec)
-- FIXME: the threads ids aren't used yet and they are a potential risk given
--	  the current handling of TCBs by the storage manager
  return $ ProcId pid [] (valV vec)
  where
    writer hdl []     = do
			  hClose hdl
			  logDebug $ 
			    "Processes: `" ++ fname ++ "': Writer finished"
    writer hdl (c:cs) = hPutChar hdl c >> writer hdl cs
    --
--    reader hdl po     = let read = do
--  FIXME: just temporary
    reader hdl po close     = let read = do
				     c <- hGetChar hdl
				     po <-- c
				     read
			in
			read `catch` \e -> do
			  unless (isEOFError e) $ ioError e
			  logDebug $ 
			    "Processes: `" ++ fname ++ "': A reader finished"
			  when close $ closePort po -- FIXME: KLUDGE
    --
    -- FIXME: the following is awful and ugly, but doing it properly requires
    --	      support for OS threads in the Haskell runtime
    waitForExitCode pid vec = do
      status <- do
		  status <- getProcessStatus False False pid
		  -- Note: With WNOHANG, waitpid() can have a return value of
		  --	   0 and still an `errno != 0'; the current
		  --	   implementation of `getProcessStatus' doesn't handle
		  --	   this case.
		  errno  <- getErrorCode
		  if errno == noChildProcess 
		    then do
		      logDebug $ 
		        "Processes: `" ++ fname ++ "': Ooops - child vanished"
		      return $ Just (Exited ExitSuccess)
		    else
		      return status
      when (isJust status) $
        logDebug $ "Processes: `" ++ fname ++ "': Terminated"
      case status of
        Nothing		      -> do
				   threadDelay 10000	-- wait 10ms
				   waitForExitCode pid vec
	Just (Exited ec     ) -> vec << ec
	Just (Terminated sig) -> vec << ExitFailure (128 + sig)
	Just _		      -> -- can't happen, as second argument = `False'
	  error "Processes.proc: Stopped?"
    --
    stdinDesc  = "`" ++ fname ++ "' stdin writer"
    stdoutDesc = "`" ++ fname ++ "' stdout reader"
    stderrDesc = "`" ++ fname ++ "' stderr reader"
    ecDesc     = "`" ++ fname ++ "' exit code poller"


-- process combinators
-- -------------------

-- pipes the output of one process into another; stderr is merged (EXPORTED)
--
(>|)     :: Proc -> Proc -> Proc
p1 >| p2  = \stdin stdout stderr -> do
	      pipeIn  <- newPort undefined
	      pipeOut <- listenToPort pipeIn
	      pid1 <- p1 stdin   pipeIn stderr
	      pid2 <- p2 pipeOut stdout stderr
	      return $ PairId pid1 pid2


-- process operators
-- -----------------

-- returns the exit status of the given process (group) (EXPORTED)
--
-- * blocks until the value is available
--
-- * in case of a `NoProcId', the availability of the unit argument implements
--   a synchronisation point
--
-- * in case of a process group, the status is `ExitSuccess' iff the exit
--   status of all processes is `ExitSuccess'; otherwise, it is the failure
--   status of any of the failed processes (it is, of course, always
--   deterministically the same, which is picked)
--
procStatus                 :: ProcId -> ExitCode
procStatus (NoProcId ())    = ExitSuccess	   -- performs synchronisation
procStatus (ProcId _ _ ec)  = ec
procStatus (PairId p1 p2 )  = case procStatus p1 of
			        ExitSuccess        -> procStatus p2
				ec@(ExitFailure _) -> ec

-- kill a process (EXPORTED)
--
procKill                  :: ProcId -> IO ()
procKill (NoProcId _    )  = return ()
procKill (ProcId pid _ _)  = signalProcess killProcess pid
procKill (PairId p1 p2  )  = procKill p1 >> procKill p2


-- auxilliary routines
-- -------------------

-- we replicate this routine from `Posix', as we need to get at the PID, which
-- the original routine does not provide us with
--
runProcess :: FilePath			    -- Command
           -> [String]			    -- Arguments
           -> Maybe [(String, String)]	    -- Environment
           -> Maybe FilePath		    -- Working directory    
           -> Maybe Handle		    -- stdin
           -> Maybe Handle		    -- stdout
           -> Maybe Handle		    -- stderr
           -> IO ProcessID
runProcess path args env dir stdin stdout stderr = do
    pid <- forkProcess
    case pid of
      Nothing  -> do 
		    doTheBusiness
		      `catch` \e -> do
				      putStrLn $ path ++ ": " ++ show e
				      exitImmediately $ ExitFailure 255
		    error "To shut the type checker up!"
      Just pid -> return pid
  where
    doTheBusiness :: IO ()
    doTheBusiness = do
        maybeChangeWorkingDirectory
        maybeDup2 0 stdin
        maybeDup2 1 stdout
        maybeDup2 2 stderr
        executeFile path True args env
	error "runProcess: This never happens"

    maybeChangeWorkingDirectory :: IO ()
    maybeChangeWorkingDirectory =
        case dir of
          Nothing -> return ()
          Just x  -> setCurrentDirectory x

    maybeDup2 :: Int -> Maybe Handle -> IO ()
    maybeDup2 dest h =
        case h of Nothing -> return ()
                  Just x  -> do
		    src <- handleToFd x
                    dupTo src (intToFd dest)
		    return ()


-- auxilliary routines for debugging
-- ---------------------------------

-- a variant of `forkIO' that installs an exception sentry if compiled with
-- debugging enabled
--
forkIODebug          :: String -> IO () -> IO ThreadId
forkIODebug desc com  = forkIO $ excSentry desc com

-- report exceptions caught by the given computation
--
-- * only in effect when compiled with debugging enabled
--
excSentry :: String	-- thread descriptor
	  -> IO ()	-- computation to watch
	  -> IO ()
{-# INLINE excSentry #-}
excSentry desc com | not debug = com
		   | otherwise = com `catch` logExc
  where
    logExc exc = do
		   putStrLn $ "Processes: Thread \"" ++ desc ++ "\": "
			      ++ show exc
		   throw exc

-- emit debugging message
--
logDebug                 :: String -> IO ()
{-# INLINE logDebug #-}
logDebug msg | not debug  = return ()
	     | otherwise  = putStrLn msg
